@@ -183,7 +183,7 @@ static const char *COMMAND_ME = "/me ";
 
 void hk_message_in(const char *bjid, const char *resname,
                    time_t timestamp, const char *msg, LmMessageSubType type,
-                   guint encrypted)
+                   guint encrypted, gboolean carbon)
 {
   int new_guy = FALSE;
   int is_groupchat = FALSE; // groupchat message
@@ -251,6 +251,7 @@ void hk_message_in(const char *bjid, const char *resname,
       { "groupchat", is_groupchat ? "true" : "false" },
       { "delayed", strdelay },
       { "error", error_msg_subtype ? "true" : "false" },
+      { "carbon", carbon ? "true" : "false" },
       { NULL, NULL },
     };
     h_result = hk_run_handlers(HOOK_PRE_MESSAGE_IN, args);
@@ -332,6 +333,9 @@ void hk_message_in(const char *bjid, const char *resname,
         }
       }
     }
+  } else if (settings_opt_get_int("roster_autolock_resource")) {
+    buddy_setactiveresource(roster_usr->data, resname);
+    scr_update_chat_status(FALSE);
   }
 
   if (error_msg_subtype) {
@@ -402,6 +406,7 @@ void hk_message_in(const char *bjid, const char *resname,
       { "attention", attention ? "true" : "false" },
       { "delayed", strdelay },
       { "error", error_msg_subtype ? "true" : "false" },
+      { "carbon", carbon ? "true" : "false" },
       { NULL, NULL },
     };
     hk_run_handlers(HOOK_POST_MESSAGE_IN, args);
@@ -546,8 +551,8 @@ void hk_statuschange(const char *bjid, const char *resname, gchar prio,
       (st_in_buf == 1 && (status == offline || oldstat == offline))) {
     // Write the status change in the buddy's buffer, only if it already exists
     if (scr_buddy_buffer_exists(bjid)) {
-      bn = g_strdup_printf("Buddy status has changed: [%c>%c] %s",
-                           imstatus2char[oldstat], imstatus2char[status],
+      bn = g_strdup_printf("Buddy status has changed: [%c>%c] [%s] %s",
+                           imstatus2char[oldstat], imstatus2char[status], rn,
                            ((status_msg) ? status_msg : ""));
       scr_write_incoming_message(bjid, bn, timestamp,
                                  HBB_PREFIX_INFO|HBB_PREFIX_NOFLAG, 0);
@@ -629,8 +634,7 @@ void hk_postconnect(void)
   scr_LogPrint(LPRINT_LOGNORM, "Running hook-post-connect...");
 
   cmdline = from_utf8(hook_command);
-  if (process_command(cmdline, TRUE) == 255)
-    mcabber_set_terminate_ui();
+  process_command(cmdline, TRUE);
 
   g_free(cmdline);
 }
@@ -656,8 +660,7 @@ void hk_predisconnect(void)
   scr_LogPrint(LPRINT_LOGNORM, "Running hook-pre-disconnect...");
 
   cmdline = from_utf8(hook_command);
-  if (process_command(cmdline, TRUE) == 255)
-    mcabber_set_terminate_ui();
+  process_command(cmdline, TRUE);
 
   g_free(cmdline);
 }
