@@ -1,7 +1,7 @@
 /*
  * utils.c      -- Various utility functions
  *
- * Copyright (C) 2005-2010 Mikael Berthe <mikael@lilotux.net>
+ * Copyright (C) 2005-2014 Mikael Berthe <mikael@lilotux.net>
  * Some of the ut_* functions are derived from Cabber debug/log code.
  * from_iso8601() comes from the Pidgin (libpurple) project.
  *
@@ -79,6 +79,7 @@ char *jidtodisp(const char *fjid)
   return alias;
 }
 
+// The caller must free the string after use.
 char *jid_get_username(const char *fjid)
 {
   char *ptr;
@@ -91,6 +92,21 @@ char *jid_get_username(const char *fjid)
   return username;
 }
 
+// The caller must free the string after use.
+char *get_servername(const char *username, const char *servername)
+{
+  char *ptr;
+  char *server;
+
+  if ((ptr = strchr(username, JID_DOMAIN_SEPARATOR)) != NULL) {
+    server = g_strdup(ptr+1);
+    return server;
+  }
+
+  return g_strdup(servername);
+}
+
+// The caller must free the string after use.
 char *compose_jid(const char *username, const char *servername,
                   const char *resource)
 {
@@ -171,7 +187,7 @@ static gboolean tracelog_create(void)
 
   fp = fopen(FName, "a");
   if (!fp) {
-    scr_LogPrint(LPRINT_NORMAL, "ERROR: Cannot open tracelog file: %s!",
+    scr_LogPrint(LPRINT_NORMAL, "ERROR: Cannot open tracelog file: %s",
                  strerror(errno));
     return FALSE;
   }
@@ -180,13 +196,17 @@ static gboolean tracelog_create(void)
   if (err || buf.st_uid != geteuid()) {
     fclose(fp);
     if (err)
-      scr_LogPrint(LPRINT_NORMAL, "ERROR: cannot stat the tracelog file: %s!",
+      scr_LogPrint(LPRINT_NORMAL, "ERROR: cannot stat the tracelog file: %s",
                    strerror(errno));
     else
       scr_LogPrint(LPRINT_NORMAL, "ERROR: tracelog file does not belong to you!");
     return FALSE;
   }
-  fchmod(fileno(fp), S_IRUSR|S_IWUSR);
+
+  if (fchmod(fileno(fp), S_IRUSR|S_IWUSR)) {
+    scr_LogPrint(LPRINT_NORMAL, "WARNING: Cannot set tracelog file permissions: %s",
+                 strerror(errno));
+  }
 
   v = mcabber_version();
   fprintf(fp, "New trace log started.  MCabber version %s\n"
@@ -196,6 +216,7 @@ static gboolean tracelog_create(void)
   return TRUE;
 }
 
+// The caller must free the string after use.
 static gchar *tracelog_level_guard(const gchar *key, const gchar *new_value)
 {
   int new_level = 0;
@@ -208,6 +229,7 @@ static gchar *tracelog_level_guard(const gchar *key, const gchar *new_value)
   return g_strdup(new_value);
 }
 
+// The caller must free the string after use.
 static gchar *tracelog_file_guard(const gchar *key, const gchar *new_value)
 {
   gchar *new_fname = NULL;
@@ -229,7 +251,7 @@ static gchar *tracelog_file_guard(const gchar *key, const gchar *new_value)
 }
 
 //  ut_init_debug()
-// Installs otpion guards before initial config file parsing.
+// Install option guards before initial config file parsing.
 void ut_init_debug(void)
 {
   DebugEnabled = 0;
@@ -250,6 +272,10 @@ void ut_write_log(unsigned int flag, const char *data)
                    strerror(errno));
       return;
     }
+
+    // Check file permissions again (it could be a new file)
+    fchmod(fileno(fp), S_IRUSR|S_IWUSR);
+
     if (fputs(data, fp) == EOF)
       scr_LogPrint(LPRINT_NORMAL, "ERROR: Cannot write to tracelog file.");
     fclose(fp);
@@ -403,7 +429,8 @@ time_t from_iso8601(const char *timestamp, int utc)
       int tzhrs, tzmins;
 
       if (*c == '.') /* dealing with precision we don't care about */
-        c += 4;
+        while (isdigit(*++c))
+          ;
 
       if ((*c == '+' || *c == '-') &&
           sscanf(c+1, "%02d:%02d", &tzhrs, &tzmins)) {
@@ -563,7 +590,7 @@ void strip_arg_special_chars(char *s)
 {
   int instring = FALSE;
   int escape = FALSE;
-  char *p;
+  char *p, *t;
 
   if (!s) return;
 
@@ -571,13 +598,17 @@ void strip_arg_special_chars(char *s)
     if (*p == '"') {
       if (!escape) {
         instring = !instring;
-        strcpy(p, p+1);
+        //memmove(p, p+1, strlen(p));
+        for (t=p; *t; t++)
+          *t = *(t+1);
         p--;
       } else
         escape = FALSE;
     } else if (*p == '\\') {
       if (!escape) {
-        strcpy(p, p+1);
+        //memmove(p, p+1, strlen(p));
+        for (t=p; *t; t++)
+          *t = *(t+1);
         p--;
       }
       escape = !escape;
@@ -676,6 +707,9 @@ char *ut_expand_tabs(const char *text)
   char *p, *q;
   guint n = 0, bc = 0;
 
+  if (!text)
+    return NULL;
+
   xtext = (char*)text;
   for (p=xtext; *p; p++)
     if (*p == '\t')
@@ -703,6 +737,53 @@ char *ut_expand_tabs(const char *text)
   return xtext;
 }
 
+//  ut_unescape_tabs_cr(text)
+// Expand CR or TAB character sequences (\n, \t) in string text.
+// If there is no CR/TAB in text, then the original pointer is returned
+// (be careful _not_ to free the pointer in this case).
+// If there are some unescaped sequences, a new string with those chars
+// replaced with real newline/tab characters is allocated; in this case
+// this is up to the caller to free this string after use.
+char *ut_unescape_tabs_cr(const char *text)
+{
+  char *xtext, *linestart;
+  char *p, *q;
+
+  if (!text)
+    return NULL;
+
+  p = g_strstr_len(text, -1, "\\n");
+  if (!p) {
+    p = g_strstr_len(text, -1, "\\t");
+    if (!p)
+      return (char*)text;
+  }
+
+  xtext = g_new(char, strlen(text) + 1);
+  p = (char*)text;
+  q = linestart = xtext;
+  do {
+    if (*p == '\\') {
+      if (*(p+1) == '\\' && (*(p+2) == 'n' || *(p+2) == 't')) {
+        // This is an escaped CR sequence
+        *q++ = '\\';
+        *q++ = 'n';
+        p += 2;
+        continue;
+      }
+      if (*(p+1) == 'n' || *(p+1) == 't') {
+        // This is a CR sequence
+        p++;
+        *q++ = (*p == 'n' ? '\n' : '\t');
+        continue;
+      }
+    }
+    *q++ = *p;
+  } while (*p++);
+
+  return xtext;
+}
+
 
 /* Cygwin's newlib does not have strcasestr() */
 /* The author of the code before the endif is
