@@ -1,8 +1,8 @@
 /*
  * xmpp.c       -- Jabber protocol handling
  *
- * Copyright (C) 2008-2010 Frank Zschockelt <mcabber@freakysoft.de>
- * Copyright (C) 2005-2010 Mikael Berthe <mikael@lilotux.net>
+ * Copyright (C) 2008-2014 Frank Zschockelt <mcabber@freakysoft.de>
+ * Copyright (C) 2005-2014 Mikael Berthe <mikael@lilotux.net>
  * Parts come from the centericq project:
  * Copyright (C) 2002-2005 by Konstantin Klyagin <konst@konst.org.ua>
  *
@@ -40,6 +40,7 @@
 #include "settings.h"
 #include "utils.h"
 #include "main.h"
+#include "carbons.h"
 
 #define RECONNECTION_TIMEOUT    60L
 
@@ -310,16 +311,13 @@ void xmpp_send_msg(const char *fjid, const char *text, int type,
   int otr_msg = 0;
 #endif
   char *barejid;
-#if defined HAVE_GPGME || defined XEP0022 || defined XEP0085
+#if defined HAVE_GPGME || defined XEP0085
   char *rname;
   GSList *sl_buddy;
 #endif
-#if defined XEP0022 || defined XEP0085
+#ifdef XEP0085
   LmMessageNode *event;
   struct xep0085 *xep85 = NULL;
-#if defined XEP0022
-  guint use_xep85 = 0;
-#endif
 #endif
   gchar *enc = NULL;
 
@@ -342,8 +340,7 @@ void xmpp_send_msg(const char *fjid, const char *text, int type,
   }
 
   barejid = jidtodisp(fjid);
-#if defined HAVE_GPGME || defined HAVE_LIBOTR || \
-    defined XEP0022 || defined XEP0085
+#if defined HAVE_GPGME || defined HAVE_LIBOTR || defined XEP0085
   rname = strchr(fjid, JID_RESOURCE_SEPARATOR);
   sl_buddy = roster_find(barejid, jidsearch, ROSTER_TYPE_USER);
 
@@ -398,7 +395,7 @@ void xmpp_send_msg(const char *fjid, const char *text, int type,
   }
 #endif // HAVE_GPGME
 
-#endif // HAVE_GPGME || defined XEP0022 || defined XEP0085
+#endif // HAVE_GPGME || defined XEP0085
 
   x = lm_message_new_with_sub_type(fjid, LM_MESSAGE_TYPE_MESSAGE, subtype);
   lm_message_node_add_child(x->node, "body",
@@ -416,6 +413,11 @@ void xmpp_send_msg(const char *fjid, const char *text, int type,
     g_free(enc);
   }
 
+  // We probably don't want Carbons for encrypted messages, since the other
+  // resources won't be able to decrypt them.
+  if (enc && carbons_enabled())
+    lm_message_node_add_child(x->node, "private", NS_CARBONS_2);
+
   // XEP-0184: Message Receipts
   if (sl_buddy && xep184 &&
       caps_has_feature(buddy_resource_getcaps(sl_buddy->data, rname),
@@ -427,16 +429,14 @@ void xmpp_send_msg(const char *fjid, const char *text, int type,
   }
   g_free(barejid);
 
-#if defined XEP0022 || defined XEP0085
+#ifdef XEP0085
   // If typing notifications are disabled, we can skip all this stuff...
   if (chatstates_disabled || type == ROSTER_TYPE_ROOM)
     goto xmpp_send_msg_no_chatstates;
 
   if (sl_buddy)
     xep85 = buddy_resource_xep85(sl_buddy->data, rname);
-#endif
 
-#ifdef XEP0085
   /* XEP-0085 5.1
    * "Until receiving a reply to the initial content message (or a standalone
    * notification) from the Contact, the User MUST NOT send subsequent chat
@@ -450,40 +450,9 @@ void xmpp_send_msg(const char *fjid, const char *text, int type,
     lm_message_node_set_attribute(event, "xmlns", NS_CHATSTATES);
     if (xep85->support == CHATSTATES_SUPPORT_UNKNOWN)
       xep85->support = CHATSTATES_SUPPORT_PROBED;
-#ifdef XEP0022
-    else
-      use_xep85 = 1;
-#endif
     xep85->last_state_sent = ROSTER_EVENT_ACTIVE;
   }
 #endif
-#ifdef XEP0022
-  /* XEP-22
-   * If the Contact supports XEP-0085, we do not use XEP-0022.
-   * If not, we try to fall back to XEP-0022.
-   */
-  if (!use_xep85) {
-    struct xep0022 *xep22 = NULL;
-    event = lm_message_node_add_child(x->node, "x", NULL);
-    lm_message_node_set_attribute(event, "xmlns", NS_EVENT);
-    lm_message_node_add_child(event, "composing", NULL);
-
-    if (sl_buddy)
-      xep22 = buddy_resource_xep22(sl_buddy->data, rname);
-    if (xep22)
-      xep22->last_state_sent = ROSTER_EVENT_ACTIVE;
-
-    // An id is mandatory when using XEP-0022.
-    if (text || subject) {
-      const gchar *msgid = lm_message_get_id(x);
-      // Let's update last_msgid_sent
-      if (xep22) {
-        g_free(xep22->last_msgid_sent);
-        xep22->last_msgid_sent = g_strdup(msgid);
-      }
-    }
-  }
-#endif
 
 xmpp_send_msg_no_chatstates:
 #ifdef WITH_DEPRECATED_STATUS_INVISIBLE
@@ -553,93 +522,21 @@ static void xmpp_send_xep85_chatstate(const char *bjid, const char *resname,
 }
 #endif
 
-#ifdef XEP0022
-//  xmpp_send_xep22_event()
-// Send a XEP-22 message event (delivered, composing...).
-static void xmpp_send_xep22_event(const char *fjid, guint type)
-{
-  LmMessage *x;
-  LmMessageNode *event;
-  const char *msgid;
-  char *rname, *barejid;
-  GSList *sl_buddy;
-  struct xep0022 *xep22 = NULL;
-  guint xep22_state;
-
-  if (!xmpp_is_online())
-    return;
-
-  rname = strchr(fjid, JID_RESOURCE_SEPARATOR);
-  barejid = jidtodisp(fjid);
-  sl_buddy = roster_find(barejid, jidsearch, ROSTER_TYPE_USER);
-  g_free(barejid);
-
-  // If we can get a resource name, we use it.  Else we use NULL,
-  // which hopefully will give us the most likely resource.
-  if (rname)
-    rname++;
-  if (sl_buddy)
-    xep22 = buddy_resource_xep22(sl_buddy->data, rname);
-
-  if (!xep22)
-    return; // XXX Maybe we could try harder (other resources?)
-
-  msgid = xep22->last_msgid_rcvd;
-
-  // For composing events (composing, active, inactive, paused...),
-  // XEP22 only has 2 states; we'll use composing and active.
-  if (type == ROSTER_EVENT_COMPOSING)
-    xep22_state = ROSTER_EVENT_COMPOSING;
-  else if (type == ROSTER_EVENT_ACTIVE ||
-           type == ROSTER_EVENT_PAUSED)
-    xep22_state = ROSTER_EVENT_ACTIVE;
-  else
-    xep22_state = 0; // ROSTER_EVENT_NONE
-
-  if (xep22_state) {
-    // Do not re-send a same event
-    if (xep22_state == xep22->last_state_sent)
-      return;
-    xep22->last_state_sent = xep22_state;
-  }
-
-  x = lm_message_new_with_sub_type(fjid, LM_MESSAGE_TYPE_MESSAGE,
-                                   LM_MESSAGE_SUB_TYPE_CHAT);
-
-  event = lm_message_node_add_child(x->node, "x", NULL);
-  lm_message_node_set_attribute(event, "xmlns", NS_EVENT);
-  if (type == ROSTER_EVENT_DELIVERED)
-    lm_message_node_add_child(event, "delivered", NULL);
-  else if (type == ROSTER_EVENT_COMPOSING)
-    lm_message_node_add_child(event, "composing", NULL);
-  lm_message_node_add_child(event, "id", msgid);
-
-  lm_connection_send(lconnection, x, NULL);
-  lm_message_unref(x);
-}
-#endif
-
 //  xmpp_send_chatstate(buddy, state)
 // Send a chatstate or event (XEP-22/85) according to the buddy's capabilities.
 // The message is sent to one of the resources with the highest priority.
-#if defined XEP0022 || defined XEP0085
+#if defined XEP0085
 void xmpp_send_chatstate(gpointer buddy, guint chatstate)
 {
   const char *bjid;
   const char *activeres;
-#ifdef XEP0085
   GSList *resources, *p_res, *p_next;
   struct xep0085 *xep85 = NULL;
-#endif
-#ifdef XEP0022
-  struct xep0022 *xep22;
-#endif
 
   bjid = buddy_getjid(buddy);
   if (!bjid) return;
   activeres = buddy_getactiveresource(buddy);
 
-#ifdef XEP0085
   /* Send the chatstate to the last resource (which should have the highest
      priority).
      If chatstate is "active", send an "active" state to all resources
@@ -665,13 +562,6 @@ void xmpp_send_chatstate(gpointer buddy, guint chatstate)
   // we don't want to send a XEP22 event.
   if (xep85 && xep85->support == CHATSTATES_SUPPORT_OK)
     return;
-#endif
-#ifdef XEP0022
-  xep22 = buddy_resource_xep22(buddy, activeres);
-  if (xep22 && xep22->support == CHATSTATES_SUPPORT_OK) {
-    xmpp_send_xep22_event(bjid, chatstate);
-  }
-#endif
 }
 #endif
 
@@ -679,13 +569,12 @@ void xmpp_send_chatstate(gpointer buddy, guint chatstate)
 //  chatstates_reset_probed(fulljid)
 // If the XEP has been probed for this contact, set it back to unknown so
 // that we probe it again.  The parameter must be a full jid (w/ resource).
-#if defined XEP0022 || defined XEP0085
+#if defined XEP0085
 static void chatstates_reset_probed(const char *fulljid)
 {
   char *rname, *barejid;
   GSList *sl_buddy;
   struct xep0085 *xep85;
-  struct xep0022 *xep22;
 
   rname = strchr(fulljid, JID_RESOURCE_SEPARATOR);
   if (!rname++)
@@ -699,12 +588,9 @@ static void chatstates_reset_probed(const char *fulljid)
     return;
 
   xep85 = buddy_resource_xep85(sl_buddy->data, rname);
-  xep22 = buddy_resource_xep22(sl_buddy->data, rname);
 
   if (xep85 && xep85->support == CHATSTATES_SUPPORT_PROBED)
     xep85->support = CHATSTATES_SUPPORT_UNKNOWN;
-  if (xep22 && xep22->support == CHATSTATES_SUPPORT_PROBED)
-    xep22->support = CHATSTATES_SUPPORT_UNKNOWN;
 }
 #endif
 
@@ -852,6 +738,7 @@ static void connection_auth_cb(LmConnection *connection, gboolean success,
   if (success) {
 
     xmpp_iq_request(NULL, NS_ROSTER);
+    xmpp_iq_request(NULL, NS_DISCO_INFO);
     xmpp_request_storage("storage:bookmarks");
     xmpp_request_storage("storage:rosternotes");
 
@@ -951,6 +838,8 @@ static void connection_close_cb(LmConnection *connection,
   if (rosternotes)
     lm_message_node_unref(rosternotes);
   rosternotes = NULL;
+  // Reset carbons
+  carbons_reset();
   // Update display
   update_roster = TRUE;
   scr_update_buddy_window();
@@ -963,123 +852,53 @@ static void connection_close_cb(LmConnection *connection,
   xmpp_setstatus(offline, NULL, mystatusmsg, TRUE);
 }
 
-static void handle_state_events(const char *from, LmMessageNode *node)
+static void handle_state_events(const char *bjid,
+                                const char *resource,
+                                LmMessageNode *node)
 {
-#if defined XEP0022 || defined XEP0085
+#if defined XEP0085
   LmMessageNode *state_ns = NULL;
-  char *rname, *bjid;
   GSList *sl_buddy;
-  guint events;
-  struct xep0022 *xep22 = NULL;
   struct xep0085 *xep85 = NULL;
-  enum {
-    XEP_none,
-    XEP_85,
-    XEP_22
-  } which_xep = XEP_none;
 
-  rname = strchr(from, JID_RESOURCE_SEPARATOR);
-  if (rname)
-    ++rname;
-  else
-    rname = (char *)from + strlen(from);
-  bjid  = jidtodisp(from);
   sl_buddy = roster_find(bjid, jidsearch, ROSTER_TYPE_USER);
-  g_free(bjid);
 
-  /* XXX Actually that's wrong, since it filters out server "offline"
-     messages (for XEP-0022).  This XEP is (almost) deprecated so
-     we don't really care. */
   if (!sl_buddy) {
     return;
   }
 
-  /* Let's see chich XEP the contact uses.  If possible, we'll use
-     XEP-85, if not we'll look for XEP-22 support. */
-  events = buddy_resource_getevents(sl_buddy->data, rname);
-
-  xep85 = buddy_resource_xep85(sl_buddy->data, rname);
-  if (xep85) {
+  /* Let's see whether the contact supports XEP0085 */
+  xep85 = buddy_resource_xep85(sl_buddy->data, resource);
+  if (xep85)
     state_ns = lm_message_node_find_xmlns(node, NS_CHATSTATES);
-    if (state_ns)
-      which_xep = XEP_85;
-  }
 
-  if (which_xep != XEP_85) { /* Fall back to XEP-0022 */
-    xep22 = buddy_resource_xep22(sl_buddy->data, rname);
-    if (xep22) {
-      state_ns = lm_message_node_find_xmlns(node, NS_EVENT);
-      if (state_ns)
-        which_xep = XEP_22;
-    }
-  }
-
-  if (!which_xep) { /* Sender does not use chat states */
+  if (!state_ns) { /* Sender does not use chat states */
     return;
   }
 
-  if (which_xep == XEP_85) { /* XEP-0085 */
-    xep85->support = CHATSTATES_SUPPORT_OK;
+  xep85->support = CHATSTATES_SUPPORT_OK;
 
-    if (!strcmp(state_ns->name, "composing")) {
-      xep85->last_state_rcvd = ROSTER_EVENT_COMPOSING;
-    } else if (!strcmp(state_ns->name, "active")) {
-      xep85->last_state_rcvd = ROSTER_EVENT_ACTIVE;
-    } else if (!strcmp(state_ns->name, "paused")) {
-      xep85->last_state_rcvd = ROSTER_EVENT_PAUSED;
-    } else if (!strcmp(state_ns->name, "inactive")) {
-      xep85->last_state_rcvd = ROSTER_EVENT_INACTIVE;
-    } else if (!strcmp(state_ns->name, "gone")) {
-      xep85->last_state_rcvd = ROSTER_EVENT_GONE;
-    }
-    events = xep85->last_state_rcvd;
-  } else {              /* XEP-0022 */
-#ifdef XEP0022
-    const char *body = lm_message_node_get_child_value(node, "body");
-    const char *msgid;
-    xep22->support = CHATSTATES_SUPPORT_OK;
-    xep22->last_state_rcvd = ROSTER_EVENT_NONE;
-
-    msgid = lm_message_node_get_attribute(node, "id");
-
-    if (lm_message_node_get_child(state_ns, "composing")) {
-      // Clear composing if the message contains a body
-      if (body)
-        events &= ~ROSTER_EVENT_COMPOSING;
-      else
-        events |= ROSTER_EVENT_COMPOSING;
-      xep22->last_state_rcvd |= ROSTER_EVENT_COMPOSING;
-
-    } else {
-      events &= ~ROSTER_EVENT_COMPOSING;
-    }
-
-    // Cache the message id
-    g_free(xep22->last_msgid_rcvd);
-    if (msgid)
-      xep22->last_msgid_rcvd = g_strdup(msgid);
-    else
-      xep22->last_msgid_rcvd = NULL;
-
-    if (lm_message_node_get_child(state_ns, "delivered")) {
-      xep22->last_state_rcvd |= ROSTER_EVENT_DELIVERED;
-
-      // Do we have to send back an ACK?
-      if (body)
-        xmpp_send_xep22_event(from, ROSTER_EVENT_DELIVERED);
-    }
-#endif
+  if (!strcmp(state_ns->name, "composing")) {
+    xep85->last_state_rcvd = ROSTER_EVENT_COMPOSING;
+  } else if (!strcmp(state_ns->name, "active")) {
+    xep85->last_state_rcvd = ROSTER_EVENT_ACTIVE;
+  } else if (!strcmp(state_ns->name, "paused")) {
+    xep85->last_state_rcvd = ROSTER_EVENT_PAUSED;
+  } else if (!strcmp(state_ns->name, "inactive")) {
+    xep85->last_state_rcvd = ROSTER_EVENT_INACTIVE;
+  } else if (!strcmp(state_ns->name, "gone")) {
+    xep85->last_state_rcvd = ROSTER_EVENT_GONE;
   }
 
-  buddy_resource_setevents(sl_buddy->data, rname, events);
-
+  buddy_resource_setevents(sl_buddy->data, resource, xep85->last_state_rcvd);
   update_roster = TRUE;
 #endif
 }
 
 static void gotmessage(LmMessageSubType type, const char *from,
-                       const char *body, const char *enc, const char *subject,
-                       time_t timestamp, LmMessageNode *node_signed)
+                       const char *body, const char *enc,
+                       const char *subject, time_t timestamp,
+                       LmMessageNode *node_signed, gboolean carbon)
 {
   char *bjid;
   const char *rname;
@@ -1181,7 +1000,7 @@ static void gotmessage(LmMessageSubType type, const char *from,
         fullbody = g_strdup_printf("[%s]\n", subject);
       body = fullbody;
     }
-    hk_message_in(bjid, rname, timestamp, body, type, encrypted);
+    hk_message_in(bjid, rname, timestamp, body, type, encrypted, carbon);
     g_free(fullbody);
   }
 
@@ -1206,6 +1025,8 @@ static LmHandlerResult handle_messages(LmMessageHandler *handler,
   const char *subject = NULL;
   time_t timestamp = 0L;
   LmMessageSubType mstype;
+  gboolean skip_process = FALSE;
+  LmMessageNode *ns_signed = NULL;
 
   mstype = lm_message_get_sub_type(m);
 
@@ -1263,18 +1084,80 @@ static LmHandlerResult handle_messages(LmMessageHandler *handler,
   if (mstype == LM_MESSAGE_SUB_TYPE_ERROR) {
     x = lm_message_node_get_child(m->node, "error");
     display_server_error(x, from);
-#if defined XEP0022 || defined XEP0085
+#ifdef XEP0085
     // If the XEP85/22 support is probed, set it back to unknown so that
     // we probe it again.
     chatstates_reset_probed(from);
 #endif
   } else {
-    handle_state_events(from, m->node);
+    handle_state_events(bjid, res, m->node);
   }
 
-  if (from && (body || subject))
+  // Check for carbons!
+  x = lm_message_node_find_xmlns(m->node, NS_CARBONS_2);
+  gboolean carbons = FALSE;
+  if (x) {
+    carbons = TRUE;
+    // Parse a message that is send to one of our other resources
+    if (!g_strcmp0(x->name, "received")) {
+      // Go 1 level deeper to the forwarded message
+      x = lm_message_node_find_xmlns(x, NS_FORWARD);
+      x = lm_message_node_get_child(x, "message");
+
+      from = lm_message_node_get_attribute(x, "from");
+      if (!from) {
+        scr_LogPrint(LPRINT_LOGNORM, "Malformed carbon copy!");
+        goto handle_messages_return;
+      }
+      g_free(bjid);
+      bjid = g_strdup(from);
+      res = strchr(bjid, JID_RESOURCE_SEPARATOR);
+      if (res) *res++ = 0;
+
+      if (body && *body && !subject && !enc)
+        ns_signed = lm_message_node_find_xmlns(x, NS_SIGNED);
+      else
+        skip_process = TRUE;
+
+      // Try to handle forwarded chat state messages
+      handle_state_events(from, res, x);
+
+      scr_LogPrint(LPRINT_DEBUG, "Received incoming carbon from <%s>", from);
+
+    } else if (!g_strcmp0(x->name, "sent")) {
+      x = lm_message_node_find_xmlns(x, NS_FORWARD);
+      x = lm_message_node_get_child(x, "message");
+
+      const char *to= lm_message_node_get_attribute(x, "to");
+      if (!to) {
+        scr_LogPrint(LPRINT_LOGNORM, "Malformed carbon copy!");
+        goto handle_messages_return;
+      }
+      g_free(bjid);
+      bjid = jidtodisp(to);
+
+      if (body && *body)
+        hk_message_out(bjid, NULL, timestamp, body, 0, NULL);
+
+      scr_LogPrint(LPRINT_DEBUG, "Received outgoing carbon for <%s>", to);
+      goto handle_messages_return;
+    }
+  } else { // Not a Carbon
+    ns_signed = lm_message_node_find_xmlns(m->node, NS_SIGNED);
+  }
+
+  // Do not process the message if some fields are missing
+  if (!from || (!body && !subject))
+    skip_process = TRUE;
+
+  if (!skip_process) {
     gotmessage(mstype, from, body, enc, subject, timestamp,
-               lm_message_node_find_xmlns(m->node, NS_SIGNED));
+               ns_signed, carbons);
+  }
+
+  // We're done if it was a Carbon forwarded message
+  if (carbons)
+    goto handle_messages_return;
 
   // Report received message if message delivery receipt was requested
   if (lm_message_node_get_child(m->node, "request") &&
@@ -1294,7 +1177,8 @@ static LmHandlerResult handle_messages(LmMessageHandler *handler,
 
   { // xep184 receipt confirmation
     LmMessageNode *received = lm_message_node_get_child(m->node, "received");
-    if (received && !g_strcmp0(lm_message_node_get_attribute(received, "xmlns"), NS_RECEIPTS)) {
+    if (received && !g_strcmp0(lm_message_node_get_attribute(received, "xmlns"),
+                               NS_RECEIPTS)) {
       char       *jid = jidtodisp(from);
       const char *id  = lm_message_node_get_attribute(received, "id");
       // This is for backward compatibility; if the remote client didn't add
@@ -1303,6 +1187,16 @@ static LmHandlerResult handle_messages(LmMessageHandler *handler,
         id = lm_message_get_id(m);
       scr_remove_receipt_flag(jid, id);
       g_free(jid);
+
+#ifdef MODULES_ENABLE
+      {
+        hk_arg_t args[] = {
+          { "jid", from },
+          { NULL, NULL },
+        };
+        hk_run_handlers("hook-mdr-received", args);
+      }
+#endif
     }
   }
 
@@ -1324,6 +1218,7 @@ static LmHandlerResult handle_messages(LmMessageHandler *handler,
     }
   }
 
+handle_messages_return:
   g_free(bjid);
   return LM_HANDLER_RESULT_REMOVE_MESSAGE;
 }
@@ -2202,6 +2097,27 @@ const char *xmpp_get_bookmark_nick(const char *bjid)
   return NULL;
 }
 
+//  xmpp_get_bookmark_password(roomjid)
+// Return the room password if it is present in a bookmark.
+const char *xmpp_get_bookmark_password(const char *bjid)
+{
+  LmMessageNode *x;
+
+  if (!bookmarks || !bjid)
+    return NULL;
+
+  // Walk through the storage bookmark tags
+  for (x = bookmarks->children ; x; x = x->next) {
+    // If the node is a conference item, check the jid.
+    if (x->name && !strcmp(x->name, "conference")) {
+      const char *fjid = lm_message_node_get_attribute(x, "jid");
+      if (fjid && !strcasecmp(bjid, fjid))
+        return lm_message_node_get_child_value(x, "password");
+    }
+  }
+  return NULL;
+}
+
 int xmpp_get_bookmark_autojoin(const char *bjid)
 {
   LmMessageNode *x;
@@ -2243,7 +2159,7 @@ GSList *xmpp_get_all_storage_bookmarks(void)
     // If the node is a conference item, let's add the note to our list.
     if (x->name && !strcmp(x->name, "conference")) {
       struct bookmark *bm_elt;
-      const char *autojoin, *name, *nick;
+      const char *autojoin, *name, *nick, *passwd;
       const char *fjid = lm_message_node_get_attribute(x, "jid");
       if (!fjid)
         continue;
@@ -2252,12 +2168,15 @@ GSList *xmpp_get_all_storage_bookmarks(void)
       autojoin = lm_message_node_get_attribute(x, "autojoin");
       nick = lm_message_node_get_child_value(x, "nick");
       name = lm_message_node_get_attribute(x, "name");
+      passwd = lm_message_node_get_child_value(x, "password");
       if (autojoin && (!strcmp(autojoin, "1") || !strcmp(autojoin, "true")))
         bm_elt->autojoin = 1;
       if (nick)
         bm_elt->nick = g_strdup(nick);
       if (name)
         bm_elt->name = g_strdup(name);
+      if (passwd)
+        bm_elt->password = g_strdup(passwd);
       sl_bookmarks = g_slist_append(sl_bookmarks, bm_elt);
     }
   }
